import pytest
from com.terraquantum.experiment.v1.experimentrun.experiment_run_pb2 import (
    ExperimentRunStatusProto,
)

from tq42.experiment_run import ExperimentRun, HardwareProto
from tq42.functional_tests.functional_test_config import FunctionalTestConfig
from tq42.utils.decorators import timeout


@pytest.fixture
def config():
    return FunctionalTestConfig()


@timeout(240)
@pytest.fixture
def circuit_storage_id(config):
    parameters = {
        "parameters": {
            "qubo": [
                0,
                1,
                0,
                0,
                0,
                0,
                0,
                1,
                1,
                0,
                0,
                0,
                0,
                0,
                1,
                0,
                0,
                0,
                0,
                1,
                0,
                0,
                0,
                0,
                0,
            ],
            "number_layers": 5,
            "steps": 25,
            "velocity": 0.05,
            "optimizer": "ADAM",
        },
        "inputs": {},
    }
    exp_run = ExperimentRun.create(
        client=config.get_client(),
        algorithm="QUENC",
        version="0.4.0",
        experiment_id=config.exp,
        compute=HardwareProto.SMALL,
        parameters=parameters,
    )

    return exp_run.poll().outputs.get("circuit").get("storage_id")


@timeout(240)
def test_circuit_runner_successfully_runs(config, circuit_storage_id):
    parameters = {
        "parameters": {"shots": 500, "backend": "CIRQ_SIMULATOR"},
        "inputs": {"circuit": {"storage_id": circuit_storage_id}},
    }

    run = ExperimentRun.create(
        client=config.get_client(),
        algorithm="CIRCUIT_RUN",
        version="0.2.0",
        experiment_id=config.exp,
        compute=HardwareProto.SMALL,
        parameters=parameters,
    )

    assert config.exp == run.data.experiment_id
    assert ExperimentRunStatusProto.QUEUED == run.data.status

    run.poll()

    assert ExperimentRunStatusProto.COMPLETED == run.data.status
    assert run.result is not None



from __future__ import annotations

from typing import Optional, List

from google.protobuf import empty_pb2
from google.protobuf.json_format import MessageToJson

from tq42.project import Project
from tq42.utils.exception_handling import handle_generic_sdk_errors

from com.terraquantum.organization.v2.organization.organization_pb2 import (
    OrganizationProto,
)
from com.terraquantum.organization.v2.organization.list_organizations_pb2 import (
    ListOrganizationsResponse,
)
from com.terraquantum.organization.v2.organization.get_organization_pb2 import (
    GetOrganizationRequest,
)

from typing import TYPE_CHECKING
from tq42.utils.pretty_list import PrettyList

# only import the stuff for type hints -> avoid circular imports
if TYPE_CHECKING:
    from tq42.client import TQ42Client


class Organization:
    """
    Reference an existing organization.

    :param client: a client instance
    :param id: the id of the existing organization
    :param data: only used internally
    """

    _client: TQ42Client
    id: str
    """ID of the organization"""
    data: OrganizationProto
    """Object containing all attributes of the organization"""

    def __init__(
        self, client: TQ42Client, id: str, data: Optional[OrganizationProto] = None
    ) -> None:
        self._client = client
        self.id = id

        if data:
            self.data = data
        else:
            self.data = self._get()

    def __repr__(self) -> str:
        return f"<Organization Id={self.id} Name={self.data.name}>"

    def __str__(self) -> str:
        return f"Organization: {MessageToJson(self.data, preserving_proto_field_name=True)}"

    @handle_generic_sdk_errors
    def _get(self) -> OrganizationProto:
        """
        Gets the information about the provided organization
        """
        get_org_request = GetOrganizationRequest(organization_id=self.id)
        res: OrganizationProto = self._client.organization_client.GetOrganization(
            request=get_org_request, metadata=self._client.metadata
        )
        return res

    @staticmethod
    def from_proto(client: TQ42Client, msg: OrganizationProto) -> Organization:
        """
        Creates organization instance from a protobuf message.

        :meta private:
        """
        return Organization(client=client, id=msg.id, data=msg)

    @handle_generic_sdk_errors
    def set(self) -> Organization:
        """
        Sets the current organization as the default organization.

        :returns: organization instance
        """
        project = Project.get_default(client=self._client, organization_id=self.id)
        if project:
            project.set()
            return self

        raise KeyError()

    @staticmethod
    @handle_generic_sdk_errors
    def get_default_org(client: TQ42Client) -> Optional[Organization]:
        """
        Gets the default organization for this user based on the default_org field

        :returns: the default organization if one is set as a default
        """
        org_list = list_all(client=client)
        if len(org_list) == 0:
            return None

        orgs_sorted = sorted(org_list, key=lambda o: o.id)
        return orgs_sorted[0]


@handle_generic_sdk_errors
def list_all(client: TQ42Client) -> List[Organization]:
    """
    List all the organizations you have permission to view.

    :param client: a client instance
    :returns: a list of all organizations
    """
    empty = empty_pb2.Empty()
    res: ListOrganizationsResponse = client.organization_client.ListOrganizations(
        request=empty, metadata=client.metadata
    )
    return PrettyList(
        [Organization.from_proto(client=client, msg=msg) for msg in res.organizations]
    )



import json
import os.path
import uuid

import pytest
from com.terraquantum.experiment.v1.experimentrun.experiment_run_pb2 import (
    ExperimentRunStatusProto,
)
from com.terraquantum.storage.v1alpha1.storage_pb2 import DatasetSensitivityProto

from tq42.dataset import Dataset
from tq42.experiment_run import ExperimentRun, HardwareProto
from tq42.functional_tests.functional_test_config import FunctionalTestConfig
from tq42.utils.decorators import timeout


@pytest.fixture
def config() -> FunctionalTestConfig:
    return FunctionalTestConfig()


@pytest.fixture
def data_storage_id(config: FunctionalTestConfig) -> str:
    data_path = os.path.join(os.path.dirname(__file__), "data.json")
    dataset = Dataset.create(
        client=config.get_client(),
        project_id=config.proj,
        name=str(uuid.uuid4()),
        description="Test dataset",
        sensitivity=DatasetSensitivityProto.SENSITIVE,
        file=data_path,
    )
    return dataset.id


@timeout(240)
def test_routing_successfully_runs(config: FunctionalTestConfig, data_storage_id: str):
    routing_params = {
        "nbGranular": 20,
        "mu": 25,
        "lambda": 40,
        "nbElite": 4,
        "nbClose": 5,
        "nbIterPenaltyManagement": 1000,
        "targetFeasible": 0.2,
        "seed": -1,
        "nbIter": 40000,
        "nbIterImpr": 25000,
        "nbIterTraces": 500,
        "timeOutput": 0.5,
        "timeLimit": 30,
        "penalty_cap": 3,
        "penalty_tw": 30,
        "penalty_dur": 20,
        "n_vehicles": 0,
        "round": 2,
    }

    run = ExperimentRun.create(
        client=config.get_client(),
        algorithm="ROUTING",
        version="0.1.4",
        experiment_id=config.exp,
        compute=HardwareProto.SMALL,
        parameters={
            "parameters": {"parameter_string": json.dumps(routing_params)},
            "inputs": {"data": {"storage_id": data_storage_id}},
        },
    )

    assert config.exp == run.data.experiment_id
    assert ExperimentRunStatusProto.QUEUED == run.data.status

    run.poll()

    assert ExperimentRunStatusProto.COMPLETED == run.data.status
    assert run.outputs is not None
    assert "solution" in run.outputs
    assert run.outputs.get("solution").get("storage_id") is not None



import asyncio
import json
from typing import Awaitable

import pandas as pd
import pytest
from com.terraquantum.experiment.v1.experimentrun.experiment_run_pb2 import (
    ExperimentRunStatusProto,
)

from tq42.channel import Ask, Tell, Channel
from tq42.experiment_run import ExperimentRun, HardwareProto
from tq42.functional_tests.functional_test_config import FunctionalTestConfig
from tq42.utils.decorators import timeout


def _mass_flow_lm(row):
    return (
        0.019745030906364664
        + 0.00013074793399258446 * row["alpha"]
        + 0.0004949044385859217 * row["beta"]
        + 0.04551741292831213 * row["d2"]
        + 0.0007926550728449819 * row["l1"]
        + 0.000271121128077387 * row["l3"] * -5.9455600147173825e-05 * row["l4"]
    )


def _tumble_lm(row):
    return (
        2.8951535034749316
        + -0.013713771116343086 * row["alpha"]
        + -0.040331795422272596 * row["beta"]
        + -0.004121931530295305 * row["d1"]
        + -1.0597227166208407 * row["d2"]
        + -0.03846219537291295 * row["l1"]
        + 0.00020282278982582672 * row["l2"]
        + -0.008908198543821947 * row["l3"]
        + 0.003848225284017235 * row["l4"]
    )


@pytest.fixture
def config() -> FunctionalTestConfig:
    return FunctionalTestConfig()


@pytest.fixture
async def channels(config) -> tuple[Channel, Channel]:
    client = config.get_client()
    mass_flow_channel = await Channel.create(client=client)
    tumble_channel = await Channel.create(client=client)
    return mass_flow_channel, tumble_channel


@timeout(240)
@pytest.mark.asyncio
async def test_optimax_successfully_runs(
    config: FunctionalTestConfig, channels: Awaitable[tuple[Channel, Channel]]
):
    mass_flow_channel, tumble_channel = await channels
    optimax_params = {
        "variable_names": ["alpha", "beta", "d1", "d2", "l1", "l2", "l3", "l4"],
        "variable_lb": [52.0, 20.0, 0.5, 0.4, 0.0, 10.0, 0.0, 75.0],
        "variable_ub": [56.0, 30.0, 5.0, 0.9, 7.0, 45.0, 25.0, 95.0],
        "objective_names": ["mass_flow", "tumble"],
        "objective_aims": ["maximize", "maximize"],
        "reference_point": [0.0, 0.0],
        "func_eval_worker_channel_ids": [
            mass_flow_channel.id,
            tumble_channel.id,
        ],
        "max_generation": 1,
        "pf_opt_lambda": 2,
        "hv_opt_repetition": 1,
        "tq_hv_opt_iterations": 3,
        "tq_hv_opt_max_rank": 2,
        "tq_hv_opt_grid": 32,
    }

    run = ExperimentRun.create(
        client=config.get_client(),
        algorithm="OPTIMAX",
        version="0.0.3",
        experiment_id=config.exp,
        compute=HardwareProto.SMALL,
        parameters={
            "parameters": {"parameter_string": json.dumps(optimax_params)},
            "inputs": {},
        },
    )

    assert config.exp == run.data.experiment_id
    assert ExperimentRunStatusProto.QUEUED == run.data.status

    def discard():
        pass

    await asyncio.gather(
        mass_flow_channel.connect(
            callback=_mass_flow_callback,
            finish_callback=discard,
            max_duration_in_sec=None,
            message_timeout_in_sec=1200,
        ),
        tumble_channel.connect(
            callback=_tumble_callback,
            finish_callback=discard,
            max_duration_in_sec=None,
            message_timeout_in_sec=1200,
        ),
    )

    run.poll()

    assert ExperimentRunStatusProto.COMPLETED == run.data.status
    assert run.result is not None

    assert "mass_flow" in run.result
    assert "tumble" in run.result


async def _mass_flow_callback(ask: Ask) -> Tell:
    mass_flow = []
    for parameter in ask.parameters:
        row = pd.Series()
        for i, variable_name in enumerate(ask.headers):
            row[variable_name] = parameter.values[i]
        mass_flow.append(float(_mass_flow_lm(row)))
    return Tell(parameters=ask.parameters, headers=ask.headers, results=mass_flow)


async def _tumble_callback(ask: Ask) -> Tell:
    tumble = []
    for parameter in ask.parameters:
        row = pd.Series()
        for i, variable_name in enumerate(ask.headers):
            row[variable_name] = parameter.values[i]
        tumble.append(float(_tumble_lm(row)))
    return Tell(parameters=ask.parameters, headers=ask.headers, results=tumble)



import asyncio
from typing import Awaitable

import numpy as np
import pytest
from com.terraquantum.experiment.v1.experimentrun.experiment_run_pb2 import (
    ExperimentRunStatusProto,
)

from tq42.channel import Ask, Tell, Channel
from tq42.experiment_run import ExperimentRun, HardwareProto
from tq42.functional_tests.functional_test_config import FunctionalTestConfig
import OptimizationTestFunctions as otf
from scipy import optimize

from tq42.utils.decorators import timeout


@pytest.fixture
def config() -> FunctionalTestConfig:
    return FunctionalTestConfig()


@pytest.fixture
async def channels(config) -> tuple[Channel, Channel]:
    client = config.get_client()
    objective_func_channel = await Channel.create(client=client)
    local_opt_channel = await Channel.create(client=client)
    return objective_func_channel, local_opt_channel


@timeout(240)
@pytest.mark.asyncio
async def test_tetra_opt_successfully_runs(
    config: FunctionalTestConfig, channels: Awaitable[tuple[Channel, Channel]]
):
    objective_func_channel, local_opt_channel = await channels
    tetra_opt_params = {
        "dimensionality": 4,
        "iteration_number": 2,
        "maximal_rank": 4,
        "points_number": 1,
        "quantization": False,
        "tolerance": 0.001,
        "lower_limits": [0, 0, 0, 0],
        "upper_limits": [9, 9, 9, 9],
        "grid": [10, 10, 10, 10],
        "objective_function_channel_id": objective_func_channel.id,
        "local_optimizer_channel_id": local_opt_channel.id,
    }

    run = ExperimentRun.create(
        client=config.get_client(),
        algorithm="TETRA_OPT",
        version="0.3.2",
        experiment_id=config.exp,
        compute=HardwareProto.SMALL,
        parameters={"parameters": tetra_opt_params, "inputs": {}},
    )

    assert config.exp == run.data.experiment_id
    assert ExperimentRunStatusProto.QUEUED == run.data.status

    def discard():
        pass

    await asyncio.gather(
        objective_func_channel.connect(
            callback=objective_function_callback,
            finish_callback=discard,
            max_duration_in_sec=None,
            message_timeout_in_sec=500,
        ),
        local_opt_channel.connect(
            callback=local_optimization_function_callback,
            finish_callback=discard,
            max_duration_in_sec=None,
            message_timeout_in_sec=500,
        ),
    )

    run.poll()

    assert ExperimentRunStatusProto.COMPLETED == run.data.status
    assert run.result is not None

    assert isinstance(run.result.get("x"), list), "Expected 'x' to be a list"
    assert all(
        isinstance(x, float) for x in run.result.get("x")
    ), "Expected 'x' to be a list of floats"
    assert isinstance(run.result.get("y"), list), "Expected 'y' to be a list"
    assert all(
        isinstance(y, float) for y in run.result.get("y")
    ), "Expected 'y' to be a list of floats"


async def objective_function_callback(ask: Ask) -> Tell:
    dim = len(ask.headers)
    func = otf.Ackley(dim)
    y = [float(func(np.array(parameter.values))) for parameter in ask.parameters]

    tell = Tell(parameters=ask.parameters, headers=ask.headers, results=y)
    return tell


async def local_optimization_function_callback(ask: Ask) -> Tell:
    dim = len(ask.headers)
    func = otf.Ackley(dim)
    results, new_x = [], []

    for parameter in ask.parameters:
        res = optimize.minimize(func, np.array(parameter.values))
        new_x.append({"values": res.x})
        results.append(float(res.fun))

    tell = Tell(
        parameters=ask.parameters,
        headers=ask.headers,
        results=results,
        candidates=new_x,
    )
    return tell
